#include "World.h"

#include <algorithm>

#include "Texture2D.h"

// Include our actor class:
#include "StaticMeshActor.h"

// This constructor constructs a world with multiple objects:
World::World()
    : lightCubeMesh(CMAKE_SOURCE_DIR "/models/cube.obj")
    , lightCubeShader(CMAKE_SOURCE_DIR "/shader/lightCube.vert", CMAKE_SOURCE_DIR "/shader/lightCube.frag")
    , runtime(0.f)
{
    // Pre-generate some random light colors:
    for(int i=0; i < 128; ++i){
        float r = (float)rand()/(float)(RAND_MAX);
        float g = (float)rand()/(float)(RAND_MAX);
        float b = (float)rand()/(float)(RAND_MAX);

        // Store light color in lightColors-list:
        lightColors.emplace_back(r,g,b);
    }

    // Override the first light color with a warmer color:
    lightColors[0] = Vec4f(1.0f, 0.8f, 0.6f) * 0.2f;
    lightColors[1] = Vec4f(0.6f, 0.8f, 1.0f) * 0.2f;
    lightColors[2] = Vec4f(0.8f, 1.0f, 0.6f) * 0.2f;
    lightColors[3] = Vec4f(0.6f, 1.0f, 0.8f) * 0.2f;

    // Load ground mesh and texture:
    Texture2D groundBaseTexture(CMAKE_SOURCE_DIR "/models/textures/StoneCladding_albedo.jpg");
    Texture2D groundNormalTexture(CMAKE_SOURCE_DIR "/models/textures/StoneCladding_normal.jpg");
    Texture2D groundMetallicTexture(CMAKE_SOURCE_DIR "/models/textures/StoneCladding_metallic.jpg");
    Texture2D groundRoughnessTexture(CMAKE_SOURCE_DIR "/models/textures/StoneCladding_roughness.jpg");
    Texture2D groundAmbientOcclusionTexture(CMAKE_SOURCE_DIR "/models/textures/StoneCladding_ao.jpg");

    Mesh groundMesh(CMAKE_SOURCE_DIR "/models/groundB.obj");
    actors.emplace_back(StaticMeshActor(
        groundMesh,
        Mat4f::translation(0,-0.225f,0),
        groundBaseTexture,
        groundNormalTexture,
        groundMetallicTexture,
        groundRoughnessTexture,
        groundAmbientOcclusionTexture
    ));

    // Load table mesh:
    Mesh tableMesh(CMAKE_SOURCE_DIR "/models/table.obj");
    actors.emplace_back(StaticMeshActor(tableMesh, Mat4f::translation(0,-0.225f,0)));

    // Load island mesh:
    Mesh islandMesh(CMAKE_SOURCE_DIR "/models/bunny.obj");
    actors.emplace_back(StaticMeshActor(islandMesh, Mat4f::translation(0,1-0.225f,0) * Mat4f::scale(0.5f), Texture2D(), Texture2D(), Texture2D(), Texture2D(), Texture2D()));

    Texture2D constrBaseTexture(CMAKE_SOURCE_DIR "/models/textures/OldRustedMetal_albedo.jpg");
    Texture2D constrMetallicTexture(CMAKE_SOURCE_DIR "/models/textures/OldRustedMetal_metallic.jpg");
    Texture2D constrRoughnessTexture(CMAKE_SOURCE_DIR "/models/textures/OldRustedMetal_roughness.jpg");

    // Load table mesh:
    Mesh constructionMesh(CMAKE_SOURCE_DIR "/models/construction.obj");
    actors.emplace_back(StaticMeshActor(constructionMesh, Mat4f::translation(0,-0.225f,0), constrBaseTexture, constrMetallicTexture, constrRoughnessTexture));

}

void World::simulate(){
    // Add 1/60 s to runtime:
    runtime += 0.01667f;
}


Vec4f getLightPos(int i, float runtime){
    if(i == 0)
        return Vec4f(1.5f, 1.25f, 0.f);
    else if(i == 1)
        return Vec4f(-1.5f, 1.25f, 0.f);
    else if(i == 2)
        return Vec4f(0.f, 1.25f, 1.5f);
    else if(i == 3)
        return Vec4f(0.f, 1.25f, -1.5f);

    float t = runtime * 0.1f + i * 75.67f;
    return Vec4f(5.f * sin(t * 1.86f), sin(t * 1.38f) * 3.f + 3, 5.f * cos(t * 1.45f));
}

void World::bindLights(Mat4f view, Shader shader, int lightNum){
    // Set the num of lights to be used in the shader:
    shader.setUniform("lightNum", lightNum);

    // Scale the intensity and the range with this factor to make the scene also bright
    // with a few light but not to bright with a lot of lights:
    float intensityfactor = 1 + pow(1 - lightNum / 100.f, 4);

    // Generate light positions and bind them to the shader:
    for(int i=0; i < 100; ++i){
        Vec4f lightPosWS = getLightPos(i, runtime);
        Vec4f lightPosCS = view * lightPosWS;

        std::string lightIdxStr = std::to_string(i);
        shader.setUniform("lights["+lightIdxStr+"].position", lightPosCS, 3);
        shader.setUniform("lights["+lightIdxStr+"].intensity", lightColors[i] * intensityfactor * 5, 3);
        shader.setUniform("lights["+lightIdxStr+"].phong_diffuse", lightColors[i] * intensityfactor * 1.9f, 3);
        shader.setUniform("lights["+lightIdxStr+"].phong_specular", lightColors[i] * intensityfactor * 0.7f, 3);
    }
}

// In this framework, we already assume that the given shader is bound.
void World::render(Mat4f view, Mat4f projection, Shader shader){
    // Now render all actors with the bound lights:
    for(StaticMeshActor& actor : actors){
        actor.render(view, projection, shader);
    }
}

void World::renderLightCubes(Mat4f view, Mat4f projection, int lightNum){
    // Generate light positions and bind them to the shader:
    for(int i=0; i < lightNum; ++i){
        Vec4f lightPosWS = getLightPos(i, runtime);

        std::string lightIdxStr = std::to_string(i);
        lightCubeShader.bind();
        lightCubeShader.setUniform("projection", projection);
        lightCubeShader.setUniform("view", view);
        lightCubeShader.setUniform("lightPosition", lightPosWS, 3);
        lightCubeShader.setUniform("lightColor", lightColors[i], 3);
        lightCubeMesh.render();
    }
}
